@@ -27,9 +27,15 @@
 #include <X11/cursorfont.h>
 #include <X11/StringDefs.h>
 #include <X11/Shell.h>
+#ifdef VMS
+#include <X11/bitmaps/dot.>
+#include <X11/bitmaps/menu16.>
+#include <X11/bitmaps/tie_fighter.>
+#else
 #include <X11/bitmaps/dot>
 #include <X11/bitmaps/menu16>
 #include <X11/bitmaps/tie_fighter>
+#endif
 
 #include <X11/Xaw/Cardinals.h>
 #include <X11/Xaw/Form.h>
@@ -44,12 +50,12 @@
 #include <X11/Xaw/Command.h>
 
 #include "Ghostview.h"
-#include "ghostview.h"
+#include "gv.h"
 #include "ps.h"
 
 extern char *getenv();
 
-static String version = "Ghostview, version 1.3";
+static String version = "Ghostview, version 1.4";
 
 static XtResource resources[] = {
     {"showTitle", "Labels", XtRBoolean, sizeof(Boolean),
@@ -59,14 +65,15 @@ static XtResource resources[] = {
     {"showLocator", "Labels", XtRBoolean, sizeof(Boolean),
      XtOffsetOf(AppResources, show_locator), XtRImmediate, (XtPointer)True},
     {"installStdCmap", "InstallStdCmap", XtRBoolean, sizeof(Boolean),
-     XtOffsetOf(AppResources, install_std_cmap), XtRImmediate, (XtPointer)True},
+     XtOffsetOf(AppResources, install_std_cmap), XtRImmediate,
+     (XtPointer)False},
     {"retainStdCmap", "RetainStdCmap", XtRBoolean, sizeof(Boolean),
      XtOffsetOf(AppResources, retain_std_cmap), XtRImmediate, (XtPointer)True},
     {"privateCmap", "PrivateCmap", XtRBoolean, sizeof(Boolean),
      XtOffsetOf(AppResources, private_cmap), XtRImmediate, (XtPointer)False},
     {"autoCenter", "AutoCenter", XtRBoolean, sizeof(Boolean),
      XtOffsetOf(AppResources, auto_center), XtRImmediate, (XtPointer)True},
-    {"horizonalMargin", "Margin", XtRInt, sizeof(int),
+    {"horizontalMargin", "Margin", XtRInt, sizeof(int),
      XtOffsetOf(AppResources, wm_horiz_margin), XtRImmediate, (XtPointer)20},
     {"verticalMargin", "Margin", XtRInt, sizeof(int),
      XtOffsetOf(AppResources, wm_vert_margin), XtRImmediate, (XtPointer)44},
@@ -90,6 +97,7 @@ static XtResource resources[] = {
      XtOffsetOf(AppResources, force_pagemedia), XtRImmediate, (XtPointer)False},
     {"swapLandscape", "SwapLandscape", XtRBoolean, sizeof(Boolean),
      XtOffsetOf(AppResources, swap_landscape), XtRImmediate, (XtPointer)False},
+#ifndef VMS
 #if defined(SVR4) || defined(SYSV) || defined(USG)
     {"printCommand", "PrintCommand", XtRString, sizeof(String),
      XtOffsetOf(AppResources, print_command), XtRImmediate, "lp"},
@@ -101,10 +109,18 @@ static XtResource resources[] = {
     {"printerVariable", "PrinterVariable", XtRString, sizeof(String),
      XtOffsetOf(AppResources, printer_variable), XtRImmediate, "PRINTER"},
 #endif
-    {"defaultPrinter", "DefaultPrinter", XtRString, sizeof(String),
-     XtOffsetOf(AppResources, default_printer), XtRImmediate, NULL},
     {"printPrompt", "PrintPrompt", XtRString, sizeof(String),
      XtOffsetOf(AppResources, print_prompt), XtRImmediate, "Printer Name:"},
+#else /* VMS */
+    {"printCommand", "PrintCommand", XtRString, sizeof(String),
+     XtOffsetOf(AppResources, print_command), XtRImmediate, "Print_dcl/Delete"},
+    {"printerVariable", "PrinterVariable", XtRString, sizeof(String),
+     XtOffsetOf(AppResources, printer_variable), XtRImmediate, "GV_PRINT_QUAL"},
+    {"printPrompt", "PrintPrompt", XtRString, sizeof(String),
+     XtOffsetOf(AppResources, print_prompt), XtRImmediate, "Print Qualifiers:"},
+#endif /* VMS */
+    {"defaultPrinter", "DefaultPrinter", XtRString, sizeof(String),
+     XtOffsetOf(AppResources, default_printer), XtRImmediate, NULL},
     {"printFail", "printFail", XtRString, sizeof(String),
      XtOffsetOf(AppResources, print_fail), XtRImmediate,
      "\"%s\" command failed."},
@@ -187,6 +203,7 @@ static XtActionsRec actions[] = {
     {"GhostviewCopyright",	gv_copyright},
     {"GhostviewQuit",		gv_quit},
     {"GhostviewOpen",		gv_open},
+    {"GhostviewReopen",		gv_reopen},
     {"GhostviewSave",		gv_save},
     {"GhostviewPrintWhole",	gv_print_whole},
     {"GhostviewPrintMarked",	gv_print_marked},
@@ -216,7 +233,7 @@ static XtActionsRec actions[] = {
 };
 
 String fallback_resources[] = {
-#   include "Ghostview.ad.h"
+#   include "app-defaults.h"
     NULL
 };
 
@@ -285,6 +302,9 @@ Boolean	special_cmap;	/* Is cmap not the Default Colormap */
 Pixel	white;		/* white pixel in above colormap */
 Pixel	black;		/* black pixel in above colormap */
 Atom	wm_delete_window;	/* Atom sent to destroy a window */
+XErrorHandler	old_Xerror;	/* standard error handler */
+Boolean	dying;		/* whether an X error caused our exit */
+XErrorEvent	bomb;	/* what the error was */
 
 XtAppContext app_con;
 AppResources app_res;
@@ -301,6 +321,7 @@ AppResources app_res;
         Widget             filebutton;
         Widget                 filemenu;
         Widget                     openbutton;
+        Widget                     reopenbutton;
         Widget                     printwholebutton;
         Widget                     printmarkedbutton;
         Widget                     savebutton;
@@ -385,8 +406,13 @@ char *argv[];
     if (argc == 2) {
 	filename = XtNewString(argv[1]);
 	if (strcmp(filename, "-")) {
+#ifdef VMS
+	    if ((psfile = fopen(argv[1], "r", "mbc=100")) == NULL) {
+#else
 	    if ((psfile = fopen(argv[1], "r")) == NULL) {
-		perror("Cannot open file");
+#endif
+		fprintf(stderr, "Cannot open ");
+		perror(argv[1]);
 		exit(1);
 	    }
 	    stat(filename, &sbuf);
@@ -394,6 +420,7 @@ char *argv[];
 	}
     }
 
+    old_Xerror = XSetErrorHandler(catch_Xerror);
     scr = DefaultScreenOfDisplay(dpy);
     wm_delete_window = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
 
@@ -695,6 +722,15 @@ char *argv[];
 	XtSetArg(args[num_args], XtNforeground, black);	num_args++;
 	XtSetArg(args[num_args], XtNbackground, white);	num_args++;
     }
+    reopenbutton = XtCreateManagedWidget("reopen", smeBSBObjectClass,
+				         filemenu, args, num_args);
+    XtAddCallback(reopenbutton, XtNcallback, reopen_file, (XtPointer)NULL);
+
+							num_args = 0;
+    if (special_cmap) {
+	XtSetArg(args[num_args], XtNforeground, black);	num_args++;
+	XtSetArg(args[num_args], XtNbackground, white);	num_args++;
+    }
     printwholebutton = XtCreateManagedWidget("printwhole", smeBSBObjectClass,
 				             filemenu, args, num_args);
     XtAddCallback(printwholebutton, XtNcallback, popup_dialog,
@@ -1051,7 +1087,7 @@ char *argv[];
     num_ghosts++;
     XtAddCallback(page, XtNcallback, track_and_zoom, (XtPointer)0);
     XtAddCallback(page, XtNdestroyCallback, destroy_ghost, (XtPointer)page);
-    XtAddCallback(page, XtNmessageCallback, message, (XtPointer)0);
+    XtAddCallback(page, XtNmessageCallback, message, (XtPointer)page);
     XtAddCallback(page, XtNoutputCallback, output, (XtPointer)0);
 							num_args = 0;
     XtSetArg(args[num_args], XtNxdpi, &default_xdpi);	num_args++;
