@@ -28,7 +28,7 @@
 #define SEEK_SET 0
 #endif
 
-#include <sys/types.h>
+#include <X11/Xos.h>
 #include <signal.h>
 #ifdef SIGNALRETURNSINT
 #define SIGVAL int
@@ -57,11 +57,15 @@
 #include <errno.h>
 /* BSD 4.3 errno.h does not declare errno */
 extern int errno;
+#ifdef VMS
+#include <perror.h>
+#else
 extern int sys_nerr;
 extern char *sys_errlist[];
+#endif
 
 #include "Ghostview.h"
-#include "ghostview.h"
+#include "gv.h"
 #include "ps.h"
 
 #ifndef max
@@ -230,6 +234,9 @@ show_page(number)
 	if (number < 0) number = 0;
     }
 
+    if (set_new_orientation(number) || set_new_pagemedia(number))
+	layout_ghostview();
+
     if (toc_text) {
 	int marker;
 	current_page = number;
@@ -243,16 +250,16 @@ show_page(number)
 	} else {
 	    GhostviewEnableInterpreter(page);
 	    GhostviewSendPS(page, psfile, doc->beginprolog,
-			    doc->endprolog, False);
+			    doc->lenprolog, False);
 	    GhostviewSendPS(page, psfile, doc->beginsetup,
-			    doc->endsetup, False);
+			    doc->lensetup, False);
 	}
 	if (doc->pageorder == DESCEND)
 	    i = (doc->numpages - 1) - current_page;
 	else
 	    i = current_page;
 	GhostviewSendPS(page, psfile, doc->pages[i].begin,
-			doc->pages[i].end, False);
+			doc->pages[i].len, False);
     } else {
 	if (!GhostviewIsInterpreterRunning(page))
 	    GhostviewEnableInterpreter(page);
@@ -363,7 +370,8 @@ setup_ghostview()
     }
 
     /* Build table of contents */
-    if (doc && !doc->epsf && doc->numpages) {
+    if (doc && (!doc->epsf && doc->numpages > 0 ||
+		 doc->epsf && doc->numpages > 1)) {
 	int maxlen = 0;
 	int i, j;
 	useful_page_labels = False;
@@ -418,6 +426,7 @@ setup_ghostview()
     }
     XtSetValues(toc, args, num_args);
 
+    XtSetSensitive(reopenbutton, (psfile != NULL));
     XtSetSensitive(printwholebutton, (psfile != NULL));
     XtSetSensitive(printmarkedbutton, (psfile != NULL));
     XtSetSensitive(savebutton, (toc_text != NULL));
@@ -913,13 +922,22 @@ print_file(name, whole_mode)
     SIGVAL (*oldsig)();
     int bytes;
     char buf[BUFSIZ];
+#ifdef VMS
+    char fnam[64], *p;
+#endif
+    Boolean failed;
     String ret_val;
 
+#ifdef VMS
+    sprintf(fnam, "sys$scratch:%s.tmp", tmpnam(NULL));
+    printer = fopen(fnam, "w");
+#else /* VMS */
     if (*name != '\0') {
 	setenv(app_res.printer_variable, name, True);
     }
     oldsig = signal(SIGPIPE, SIG_IGN);
     printer = popen(app_res.print_command, "w");
+#endif /* VMS */
     if (toc_text && !whole_mode) {
 	pscopydoc(printer);
     } else {
@@ -928,13 +946,21 @@ print_file(name, whole_mode)
 	    bytes = write(fileno(printer), buf, bytes);
 	fclose(psfile);
     }
-    if (pclose(printer)) {
+#ifdef VMS
+    sprintf(buf, "%s %s %s", app_res.print_command, name, fnam);
+    failed = fclose(printer) != 0 || system(buf) != 1;
+#else /* VMS */
+    failed = pclose(printer) != 0;
+#endif /* VMS */
+    if (failed) {
 	sprintf(buf, app_res.print_fail, app_res.print_command);
 	ret_val = XtNewString(buf);
     } else {
 	ret_val = NULL;
     }
+#ifndef VMS
     signal(SIGPIPE, oldsig);
+#endif /* VMS */
     return(ret_val);
 }
 
@@ -1008,12 +1034,12 @@ pscopydoc(fp)
 	else
 	    j = i;
 	if (toc_text[toc_entry_length * j] == '*') {
-	    comment = pscopyuntil(psfile, fp, doc->pages[j].begin,
-				  doc->pages[j].end, "%%Page:");
+	    comment = pscopyuntil(psfile, fp, doc->pages[i].begin,
+				  doc->pages[i].end, "%%Page:");
 	    fprintf(fp, "%%%%Page: %s %d\n",
-		    doc->pages[j].label, page++);
+		    doc->pages[i].label, page++);
 	    free(comment);
-	    pscopy(psfile, fp, -1, doc->pages[j].end);
+	    pscopy(psfile, fp, -1, doc->pages[i].end);
 	}
     }
 
@@ -1438,3 +1464,20 @@ new_file(number)
     if (set_new_pagemedia(number)) layout_changed = True;
     if (layout_changed) layout_ghostview();
 }
+
+/* Catch X errors die gracefully if one occurs */
+int
+catch_Xerror(dpy, err)
+    Display *dpy;
+    XErrorEvent *err;
+{
+    if (err->error_code == BadImplementation) {
+	old_Xerror(dpy, err);
+	return 0;
+    }
+    if (dying) return 0;
+    dying = True;
+    bomb = *err;
+    XtDestroyWidget(toplevel);
+    return 0;
+}
