@@ -30,9 +30,14 @@
 #ifndef BUFSIZ
 #define BUFSIZ 1024
 #endif
+#include <ctype.h>
 #include <X11/Xos.h>		/* #includes the appropriate <string.h> */
 #include "ps.h"
 
+#ifdef BSD4_2
+#define memset(a,b,c) bzero(a,c)
+#endif
+
 /* length calculates string length at compile time */
 /* can only be used with character constants */
 #define length(a) (sizeof(a)-1)
@@ -62,14 +67,8 @@ struct documentmedia papersizes[] = {
 
 
 static char *readline();
-	/* Position of file just before the current line was read */
-static long file_position;
-
 static char *gettextline();
 static char *gettext();
-	/* Pointer to first character after text returned by gettext */
-static char *next_char;
-
 static int  isblank();
 
 /*
@@ -160,7 +159,7 @@ static int  isblank();
  *      Next each page starts explicitly with %%Page and ends implicitly with
  *      %%Page or %%Trailer or %%EOF.  The following comments are recognized:
  *
- *      %%Page: <text>
+ *      %%Page: <text> <uint>
  *      %%PageBoundingBox: <int> <int> <int> <int>|(atend)
  *      %%PageOrientation: Portrait|Landscape
  *      %%PageMedia: <text>
@@ -180,7 +179,7 @@ static int  isblank();
  *
  *  + A DC-3 received severe damage to one of its wings.  The wing was a total
  *    loss.  There was no replacement readily available, so the mechanic
- *    install a wing from a DC-2.
+ *    installed a wing from a DC-2.
  */
 
 struct document *
@@ -197,13 +196,21 @@ psscan(file)
     int preread;		/* flag which tells the readline isn't needed */
     int i;
     unsigned int maxpages = 0;
+    unsigned int nextpage = 1;	/* Next expected page */
+    unsigned int thispage;
+    int ignore = 0;		/* whether to ignore page ordinals */
+    char *label;
     char line[PSLINELENGTH];	/* 255 characters + 1 newline + 1 NULL */
     char text[PSLINELENGTH];	/* Temporary storage for text */
+    long position;		/* Position of the current line */
+    unsigned int line_len; 	/* Length of the current line */
+    unsigned int section_len;	/* Place to accumulate the section length */
+    char *next_char;		/* 1st char after text returned by gettext() */
     char *cp;
     struct documentmedia *dmp;
 
     rewind(file);
-    if (readline(line, sizeof line, file) == NULL) {
+    if (readline(line, sizeof line, file, &position, &line_len) == NULL) {
 	fprintf(stderr, "Warning: empty file.\n");
 	return(NULL);
     }
@@ -219,17 +226,20 @@ psscan(file)
 	memset(doc, 0, sizeof(struct document));
 	sscanf(line, "%*s %s", text);
 	doc->epsf = iscomment(text, "EPSF-");
-	doc->beginheader = file_position;
+	doc->beginheader = position;
+	section_len = line_len;
     } else {
 	return(NULL);
     }
 
     preread = 0;
-    while (preread || readline(line, sizeof line, file)) {
+    while (preread || readline(line, sizeof line, file, &position, &line_len)) {
+	if (!preread) section_len += line_len;
 	preread = 0;
 	if (iscomment(line, "%%EndComments") ||
 	    line[0] != '%' ||
-	    (line[1] == ' ' || line[1] == '\t' || line[1] == '\n')) {
+	    (!isprint(line[1]) || line[1] == ' ' ||
+	     line[1] == '\t' || line[1] == '\n')) {
 	    break;
 	} else if (!iscomment(line, "%%")) {
 	    /* Do nothing */
@@ -333,7 +343,8 @@ psscan(file)
 		fprintf(stderr, "Fatal Error: Dynamic memory exhausted.\n");
 		exit(-1);
 	    }
-	    doc->media[0].name = gettext(line+length("%%DocumentMedia:"));
+	    doc->media[0].name = gettext(line+length("%%DocumentMedia:"),
+					 &next_char);
 	    if (doc->media[0].name != NULL) {
 		if (sscanf(next_char, "%f %f", &w, &h) == 2) {
 		    doc->media[0].width = w + 0.5;
@@ -345,8 +356,9 @@ psscan(file)
 		    free(doc->media[0].name);
 	    }
 	    preread=1;
-	    while (readline(line, sizeof line, file) &&
+	    while (readline(line, sizeof line, file, &position, &line_len) &&
 		   iscomment(line, "%%+")) {
+		section_len += line_len;
 		doc->media = (struct documentmedia *)
 			     realloc(doc->media,
 				     (doc->nummedia+1)*
@@ -355,7 +367,8 @@ psscan(file)
 		    fprintf(stderr, "Fatal Error: Dynamic memory exhausted.\n");
 		    exit(-1);
 		}
-		doc->media[doc->nummedia].name = gettext(line+length("%%+"));
+		doc->media[doc->nummedia].name = gettext(line+length("%%+"),
+							 &next_char);
 		if (doc->media[doc->nummedia].name != NULL) {
 		    if (sscanf(next_char, "%f %f", &w, &h) == 2) {
 			doc->media[doc->nummedia].width = w + 0.5;
@@ -367,6 +380,7 @@ psscan(file)
 			free(doc->media[doc->nummedia].name);
 		}
 	    }
+	    section_len += line_len;
 	    if (doc->nummedia != 0) doc->default_page_media = doc->media;
 	} else if (doc->nummedia == NONE &&
 		   iscomment(line+2, "DocumentPaperSizes:")) {
@@ -377,7 +391,8 @@ psscan(file)
 		fprintf(stderr, "Fatal Error: Dynamic memory exhausted.\n");
 		exit(-1);
 	    }
-	    doc->media[0].name = gettext(line+length("%%DocumentPaperSizes:"));
+	    doc->media[0].name = gettext(line+length("%%DocumentPaperSizes:"),
+					 &next_char);
 	    if (doc->media[0].name != NULL) {
 		doc->media[0].width = 0;
 		doc->media[0].height = 0;
@@ -406,7 +421,7 @@ psscan(file)
 		else
 		    free(doc->media[0].name);
 	    }
-	    while (cp = gettext(next_char)) {
+	    while (cp = gettext(next_char, &next_char)) {
 		doc->media = (struct documentmedia *)
 			     realloc(doc->media,
 				     (doc->nummedia+1)*
@@ -446,10 +461,11 @@ psscan(file)
 		    free(doc->media[doc->nummedia].name);
 	    }
 	    preread=1;
-	    while (readline(line, sizeof line, file) &&
+	    while (readline(line, sizeof line, file, &position, &line_len) &&
 		   iscomment(line, "%%+")) {
+		section_len += line_len;
 		next_char = line + length("%%+");
-		while (cp = gettext(next_char)) {
+		while (cp = gettext(next_char, &next_char)) {
 		    doc->media = (struct documentmedia *)
 				 realloc(doc->media,
 					 (doc->nummedia+1)*
@@ -480,33 +496,50 @@ psscan(file)
 			free(doc->media[doc->nummedia].name);
 		}
 	    }
+	    section_len += line_len;
 	    if (doc->nummedia != 0) doc->default_page_media = doc->media;
 	}
     }
 
-    if (iscomment(line, "%%EndComments")) readline(line, sizeof line, file);
-    doc->endheader = file_position;
+    if (iscomment(line, "%%EndComments")) {
+	readline(line, sizeof line, file, &position, &line_len);
+	section_len += line_len;
+    }
+    doc->endheader = position;
+    doc->lenheader = section_len - line_len;
 
     /* Optional Preview comments for encapsulated PostScript files */ 
 
-    while (isblank(line) && readline(line, sizeof line, file)) {}
+    while (isblank(line) &&
+	   readline(line, sizeof line, file, &position, &line_len)) {
+    }
 
     if (doc->epsf && iscomment(line, "%%BeginPreview")) {
-	doc->beginpreview = file_position;
-	while (readline(line, sizeof line, file) &&
-	       !iscomment(line, "%%EndPreview")) {}
-	readline(line, sizeof line, file);
-	doc->endpreview = file_position;
+	doc->beginpreview = position;
+	section_len = line_len;
+	while (readline(line, sizeof line, file, &position, &line_len) &&
+	       !iscomment(line, "%%EndPreview")) {
+	    section_len += line_len;
+	}
+	section_len += line_len;
+	readline(line, sizeof line, file, &position, &line_len);
+	section_len += line_len;
+	doc->endpreview = position;
+	doc->lenpreview = section_len - line_len;
     }
 
     /* Page Defaults for Version 3.0 files */
 
-    while (isblank(line) && readline(line, sizeof line, file)) {}
+    while (isblank(line) &&
+	   readline(line, sizeof line, file, &position, &line_len)) {
+    }
 
     if (iscomment(line, "%%BeginDefaults")) {
-	doc->begindefaults = file_position;
-	while (readline(line, sizeof line, file) &&
+	doc->begindefaults = position;
+	section_len = line_len;
+	while (readline(line, sizeof line, file, &position, &line_len) &&
 	       !iscomment(line, "%%EndDefaults")) {
+	    section_len += line_len;
 	    if (!iscomment(line, "%%")) {
 		/* Do nothing */
 	    } else if (doc->default_page_orientation == NONE &&
@@ -519,7 +552,7 @@ psscan(file)
 		}
 	    } else if (page_media_set == NONE &&
 		       iscomment(line+2, "PageMedia:")) {
-		cp = gettext(line+length("%%PageMedia:"));
+		cp = gettext(line+length("%%PageMedia:"), NULL);
 		for (dmp = doc->media, i=0; i<doc->nummedia; i++, dmp++) {
 		    if (strcmp(cp, dmp->name) == 0) {
 			doc->default_page_media = dmp;
@@ -557,44 +590,59 @@ psscan(file)
 		}
 	    }
 	}
-	readline(line, sizeof line, file);
-	doc->enddefaults = file_position;
+	section_len += line_len;
+	readline(line, sizeof line, file, &position, &line_len);
+	section_len += line_len;
+	doc->enddefaults = position;
+	doc->lendefaults = section_len - line_len;
     }
 
     /* Document Prolog */
 
-    while (isblank(line) && readline(line, sizeof line, file)) {}
+    while (isblank(line) &&
+	   readline(line, sizeof line, file, &position, &line_len)) {
+    }
 
     if (!iscomment(line, "%%BeginSetup") && !iscomment(line, "%%Page:") &&
 	!iscomment(line, "%%Trailer") && !iscomment(line, "%%EOF")) {
-	doc->beginprolog = file_position;
+	doc->beginprolog = position;
+	section_len = line_len;
 	preread = 1;
 
-	while ((preread || readline(line, sizeof line, file)) &&
+	while ((preread ||
+		readline(line, sizeof line, file, &position, &line_len)) &&
 	       !iscomment(line, "%%EndProlog") &&
 	       !iscomment(line, "%%BeginSetup") &&
 	       !iscomment(line, "%%Page:") && !iscomment(line, "%%Trailer") &&
 	       !iscomment(line, "%%EOF")) {
+	    if (!preread) section_len += line_len;
 	    preread = 0;
-
 	}
+	section_len += line_len;
 	if (iscomment(line, "%%EndProlog")) {
-	    readline(line, sizeof line, file);
+	    readline(line, sizeof line, file, &position, &line_len);
+	    section_len += line_len;
 	}
-	doc->endprolog = file_position;
+	doc->endprolog = position;
+	doc->lenprolog = section_len - line_len;
     }
 
     /* Document Setup,  Page Defaults found here for Version 2 files */
 
-    while (isblank(line) && readline(line, sizeof line, file)) {}
+    while (isblank(line) &&
+	   readline(line, sizeof line, file, &position, &line_len)) {
+    }
 
     if (!iscomment(line, "%%Page:") && !iscomment(line, "%%Trailer") &&
 	!iscomment(line, "%%EOF")) {
-	doc->beginsetup = file_position;
+	doc->beginsetup = position;
+	section_len = line_len;
 	preread = 1;
-	while ((preread || readline(line, sizeof line, file)) &&
+	while ((preread ||
+		readline(line, sizeof line, file, &position, &line_len)) &&
 	       !iscomment(line, "%%EndSetup") && !iscomment(line, "%%Page:") &&
 	       !iscomment(line, "%%Trailer") && !iscomment(line, "%%EOF")) {
+	    if (!preread) section_len += line_len;
 	    preread = 0;
 	    if (!iscomment(line, "%%")) {
 		/* Do nothing */
@@ -608,7 +656,7 @@ psscan(file)
 		}
 	    } else if (page_media_set == NONE &&
 		       iscomment(line+2, "PaperSize:")) {
-		cp = gettext(line+length("%%PaperSize:"));
+		cp = gettext(line+length("%%PaperSize:"), NULL);
 		for (dmp = doc->media, i=0; i<doc->nummedia; i++, dmp++) {
 		    /* Note: Paper size comment uses down cased paper size
 		     * name.  Case insensitive compares are only used for
@@ -650,16 +698,22 @@ psscan(file)
 		}
 	    }
 	}
+	section_len += line_len;
 	if (iscomment(line, "%%EndSetup")) {
-	    readline(line, sizeof line, file);
+	    readline(line, sizeof line, file, &position, &line_len);
+	    section_len += line_len;
 	}
-	doc->endsetup = file_position;
+	doc->endsetup = position;
+	doc->lensetup = section_len - line_len;
     }
 
     /* Individual Pages */
 
-    while (isblank(line) && readline(line, sizeof line, file)) {}
+    while (isblank(line) &&
+	   readline(line, sizeof line, file, &position, &line_len)) {
+    }
 
+newpage:
     while (iscomment(line, "%%Page:")) {
 	if (maxpages == 0) {
 	    maxpages = 1;
@@ -670,6 +724,17 @@ psscan(file)
 		exit(-1);
 	    }
 	}
+	label = gettext(line+length("%%Page:"), &next_char);
+	if (sscanf(next_char, "%d", &thispage) != 1) thispage = 0;
+	if (nextpage == 1) {
+	    ignore = thispage != 1;
+	}
+	if (!ignore && thispage != nextpage) {
+	    free(label);
+	    doc->numpages--;
+	    goto continuepage;
+	}
+	nextpage++;
 	if (doc->numpages == maxpages) {
 	    maxpages++;
 	    doc->pages = (struct page *)
@@ -681,11 +746,14 @@ psscan(file)
 	}
 	memset(&(doc->pages[doc->numpages]), 0, sizeof(struct page));
 	page_bb_set = NONE;
-	doc->pages[doc->numpages].label = gettext(line+length("%%Page:"));
-	doc->pages[doc->numpages].begin = file_position;
-	while (readline(line, sizeof line, file) &&
+	doc->pages[doc->numpages].label = label;
+	doc->pages[doc->numpages].begin = position;
+	section_len = line_len;
+continuepage:
+	while (readline(line, sizeof line, file, &position, &line_len) &&
 	       !iscomment(line, "%%Page:") && !iscomment(line, "%%Trailer") &&
 	       !iscomment(line, "%%EOF")) {
+	    section_len += line_len;
 	    if (!iscomment(line, "%%")) {
 		/* Do nothing */
 	    } else if (doc->pages[doc->numpages].orientation == NONE &&
@@ -698,7 +766,7 @@ psscan(file)
 		}
 	    } else if (doc->pages[doc->numpages].media == NULL &&
 		       iscomment(line+2, "PageMedia:")) {
-		cp = gettext(line+length("%%PageMedia:"));
+		cp = gettext(line+length("%%PageMedia:"), NULL);
 		for (dmp = doc->media, i=0; i<doc->nummedia; i++, dmp++) {
 		    if (strcmp(cp, dmp->name) == 0) {
 			doc->pages[doc->numpages].media = dmp;
@@ -708,7 +776,7 @@ psscan(file)
 		free(cp);
 	    } else if (doc->pages[doc->numpages].media == NULL &&
 		       iscomment(line+2, "PaperSize:")) {
-		cp = gettext(line+length("%%PaperSize:"));
+		cp = gettext(line+length("%%PaperSize:"), NULL);
 		for (dmp = doc->media, i=0; i<doc->nummedia; i++, dmp++) {
 		    /* Note: Paper size comment uses down cased paper size
 		     * name.  Case insensitive compares are only used for
@@ -759,20 +827,33 @@ psscan(file)
 		}
 	    }
 	}
-	doc->pages[doc->numpages].end = file_position;
+	section_len += line_len;
+	doc->pages[doc->numpages].end = position;
+	doc->pages[doc->numpages].len = section_len - line_len;
 	doc->numpages++;
     }
 
     /* Document Trailer */
 
-    doc->begintrailer = file_position;
+    doc->begintrailer = position;
+    section_len = line_len;
 
     preread = 1;
-    while ((preread || readline(line, sizeof line, file)) &&
+    while ((preread ||
+	    readline(line, sizeof line, file, &position, &line_len)) &&
 	   !iscomment(line, "%%EOF")) {
+	if (!preread) section_len += line_len;
 	preread = 0;
 	if (!iscomment(line, "%%")) {
 	    /* Do nothing */
+	} else if (iscomment(line+2, "Page:")) {
+	    free(gettext(line+length("%%Page:"), &next_char));
+	    if (sscanf(next_char, "%d", &thispage) != 1) thispage = 0;
+	    if (!ignore && thispage == nextpage) {
+		doc->pages[doc->numpages-1].end = position;
+		doc->pages[doc->numpages-1].len += section_len - line_len;
+		goto newpage;
+	    }
 	} else if (bb_set == ATEND && iscomment(line+2, "BoundingBox:")) {
 	    if (sscanf(line+length("%%BoundingBox:"), "%d %d %d %d",
 		       &(doc->boundingbox[LLX]),
@@ -823,10 +904,31 @@ psscan(file)
 	    }
 	}
     }
+    section_len += line_len;
     if (iscomment(line, "%%EOF")) {
-	readline(line, sizeof line, file);
+	readline(line, sizeof line, file, &position, &line_len);
+	section_len += line_len;
+    }
+    doc->endtrailer = position;
+    doc->lentrailer = section_len - line_len;
+
+    section_len = line_len;
+    preread = 1;
+    while (preread ||
+	   readline(line, sizeof line, file, &position, &line_len)) {
+	if (!preread) section_len += line_len;
+	preread = 0;
+	if (iscomment(line, "%%Page:")) {
+	    free(gettext(line+length("%%Page:"), &next_char));
+	    if (sscanf(next_char, "%d", &thispage) != 1) thispage = 0;
+	    if (!ignore && thispage == nextpage) {
+		doc->pages[doc->numpages-1].end = position;
+		doc->pages[doc->numpages-1].len += doc->lentrailer +
+						   section_len - line_len;
+		goto newpage;
+	    }
+	}
     }
-    doc->endtrailer = file_position;
     return doc;
 }
 
@@ -869,7 +971,7 @@ gettextline(line)
 
     while (*line && (*line == ' ' || *line == '\t')) line++;
     if (*line == '(') {
-	return gettext(line);
+	return gettext(line, NULL);
     } else {
 	if (strlen(line) == 0) return NULL;
 	cp = (char *) malloc(strlen(line));
@@ -886,12 +988,12 @@ gettextline(line)
 /*
  *	gettext -- return the next text string on the line.
  *		   return NULL if nothing is present.
- *                 NOTE: next_char is modified by this routine.
  */
 
 static char *
-gettext(line)
+gettext(line, next_char)
     char *line;
+    char **next_char;
 {
     char text[PSLINELENGTH];	/* Temporary storage for text */
     char *cp;
@@ -962,7 +1064,7 @@ gettext(line)
 	    *cp++ = *line++;
     }
     *cp = '\0';
-    next_char = line;
+    if (next_char) *next_char = line;
     if (!quoted && strlen(text) == 0) return NULL;
     cp = (char *) malloc(strlen(text)+1);
     if (cp == NULL) {
@@ -980,56 +1082,118 @@ gettext(line)
  *		    comments.)
  *		    Also, skip over included documents (as indicated by
  *		    %%BeginDocument/%%EndDocument comments.)
- *		    Note: file_position is modified by this routine.
  */
 
 static char *
-readline(line, size, fp)
+readline(line, size, fp, position, line_len)
     char *line;
     int size;
     FILE *fp;
+    long *position;
+    unsigned int *line_len;
 {
     char text[PSLINELENGTH];	/* Temporary storage for text */
+    char save[PSLINELENGTH];	/* Temporary storage for text */
     char *cp;
     unsigned int num;
+    unsigned int nbytes;
     int i;
     char buf[BUFSIZ];
-    long position;
 
-    position = ftell(fp);
+    if (position) *position = ftell(fp);
     cp = fgets(line, size, fp);
+    *line_len = cp ? strlen(line) : 0;
     if (iscomment(line, "%%BeginDocument:")) {
-	while (readline(line, size, fp) && !iscomment(line, "%%EndDocument")) {}
-	cp = readline(line, size, fp);
+	strcpy(save, line+7);
+	while (readline(line, size, fp, NULL, &nbytes) &&
+	       !iscomment(line, "%%EndDocument")) {
+	    *line_len += nbytes;
+	}
+	*line_len += nbytes;
+	strcpy(line+2, save);
+    } else if (iscomment(line, "%%BeginFeature:")) {
+	strcpy(save, line+7);
+	while (readline(line, size, fp, NULL, &nbytes) &&
+	       !iscomment(line, "%%EndFeature")) {
+	    *line_len += nbytes;
+	}
+	*line_len += nbytes;
+	strcpy(line+2, save);
+    } else if (iscomment(line, "%%BeginFile:")) {
+	strcpy(save, line+7);
+	while (readline(line, size, fp, NULL, &nbytes) &&
+	       !iscomment(line, "%%EndFile")) {
+	    *line_len += nbytes;
+	}
+	*line_len += nbytes;
+	strcpy(line+2, save);
+    } else if (iscomment(line, "%%BeginFont:")) {
+	strcpy(save, line+7);
+	while (readline(line, size, fp, NULL, &nbytes) &&
+	       !iscomment(line, "%%EndFont")) {
+	    *line_len += nbytes;
+	}
+	*line_len += nbytes;
+	strcpy(line+2, save);
+    } else if (iscomment(line, "%%BeginProcSet:")) {
+	strcpy(save, line+7);
+	while (readline(line, size, fp, NULL, &nbytes) &&
+	       !iscomment(line, "%%EndProcSet")) {
+	    *line_len += nbytes;
+	}
+	*line_len += nbytes;
+	strcpy(line+2, save);
+    } else if (iscomment(line, "%%BeginResource:")) {
+	strcpy(save, line+7);
+	while (readline(line, size, fp, NULL, &nbytes) &&
+	       !iscomment(line, "%%EndResource")) {
+	    *line_len += nbytes;
+	}
+	*line_len += nbytes;
+	strcpy(line+2, save);
     } else if (iscomment(line, "%%BeginData:")) {
 	text[0] = '\0';
+	strcpy(save, line+7);
 	if (sscanf(line+length("%%BeginData:"), "%d %*s %s", &num, text) >= 1) {
 	    if (strcmp(text, "Lines") == 0) {
 		for (i=0; i < num; i++) {
 		    cp = fgets(line, size, fp);
+		    *line_len += cp ? strlen(line) : 0;
 		}
 	    } else {
 		while (num > BUFSIZ) {
 		    fread(buf, sizeof (char), BUFSIZ, fp);
+		    *line_len += BUFSIZ;
 		    num -= BUFSIZ;
 		}
 		fread(buf, sizeof (char), num, fp);
+		*line_len += num;
 	    }
 	}
-	while (readline(line, size, fp) && !iscomment(line, "%%EndData")) {}
-	cp = readline(line, size, fp);
+	while (readline(line, size, fp, NULL, &nbytes) &&
+	       !iscomment(line, "%%EndData")) {
+	    *line_len += nbytes;
+	}
+	*line_len += nbytes;
+	strcpy(line+2, save);
     } else if (iscomment(line, "%%BeginBinary:")) {
+	strcpy(save, line+7);
 	if(sscanf(line+length("%%BeginBinary:"), "%d", &num) == 1) {
 	    while (num > BUFSIZ) {
 		fread(buf, sizeof (char), BUFSIZ, fp);
+		*line_len += BUFSIZ;
 		num -= BUFSIZ;
 	    }
 	    fread(buf, sizeof (char), num, fp);
+	    *line_len += num;
 	}
-	while (readline(line, size, fp) && !iscomment(line, "%%EndBinary")) {}
-	cp = readline(line, size, fp);
+	while (readline(line, size, fp, NULL, &nbytes) &&
+	       !iscomment(line, "%%EndBinary")) {
+	    *line_len += nbytes;
+	}
+	*line_len += nbytes;
+	strcpy(line+2, save);
     }
-    file_position = position;
     return cp;
 }
 
